"""Authentication module for DHL SDK

This module provides functionality for handling 
API Key Authentication. It allows you to securely 
manage and use API keys for authenticating requests 
to the DataHowLab API. 

The module includes a class, `APIKeyAuthentication`, that can
be used to obtain and include API Key authorization headers in 
your requests. It also offers the flexibility to either pass 
an API key as an argument or retrieve it from environment variables.

Classes:
    - APIKeyAuthentication: Manages API Key authentication and provides 
      methods for generating authorization headers using the API Key.

Usage:
    To use this module, create an instance of the `APIKeyAuthentication`
    class, passing an API key as an argument or allowing it 
    to retrieve the API key from environment variables. 
    You can then call the `get_headers` method to obtain the authorization 
    headers, which can be added to your requests.

"""

import os
from typing import Optional


class APIKeyAuthentication:
    # pylint: disable=too-few-public-methods
    """
    API Key Authentication

    The API Key can be provided as an argument or retrieved from the
    environment variable defined as 'DHL_API_KEY'
    """

    def __init__(self, api_key: Optional[str] = None):
        """
        Parameters
        ----------
        api_key : str, optional
            API Key to use to authenticate with the DHL API

        Returns
        -------
        NoneType
            None
        """
        self.api_key = self._get_api_key(api_key)

    def get_headers(self) -> dict[str, str]:
        """Get the authorization headers to add to the
        requests using the API Key.

        Returns
        -------
        dict
            Authorization Headers for the request to the API
        """
        return {"Authorization": f"ApiKey {self.api_key}"}

    def _get_api_key(self, api_key: Optional[str] = None) -> str:
        """Get the API Key from the environment variables
        if not given.

        Parameters
        ----------
        api_key : str, optional
            API Key to use, by default None

        Returns
        -------
        str
            API Key
        """

        if api_key is None:
            api_key = os.environ.get("DHL_API_KEY", None)

        if api_key is None:
            raise KeyError("DHL API Key not found")

        return api_key



"""This module contains generic utility functions used in the SDK
"""

import urllib.parse as urlparse
from datetime import datetime
from functools import reduce
from typing import Optional, Union

import numpy as np
from pydantic import BaseModel, Field, model_validator

from dhl_sdk.crud import Client

Predictions = dict[str, dict[str, list[float]]]

PRODUCTS_URL = "api/db/v2/products"
RECIPES_URL = "api/db/v2/recipes"
FILES_URL = "api/db/v2/files"
EXPERIMENTS_URL = "api/db/v2/experiments"
VARIABLES_URL = "api/db/v2/variables"
GROUPS_URL = "api/db/v2/groups"
PROJECTS_URL = "api/db/v2/projects"
DATASETS_URL = "api/db/v2/datasets"
MODELS_URL = "api/db/v2/pipelineJobs"
TEMPLATES_URL = "api/db/v2/pipelineJobTemplates"
PREDICT_URL = "api/pipeline/v1/predictors"


class VariableGroupCodes:
    """Singleton class to store the variable group codes"""

    _instance = None
    variable_group_codes = {}

    def __new__(cls, client):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._initialize(client)
        return cls._instance

    def _initialize(self, client: Client):
        group_codes = {}
        groups = client.get(GROUPS_URL).json()

        for group in groups:
            group_codes[group["name"]] = (group["id"], group["code"])

        self.variable_group_codes = group_codes

    def get_variable_group_codes(self) -> dict[str, tuple[str, str]]:
        """Returns the variable group codes saved in singleton"""
        return self.variable_group_codes


class Instance(BaseModel):
    """Pydantic class representing the Instance
    It is used to type check the request"""

    timestamps: Optional[list[float]] = Field(default=None, alias="timestamps")
    sample_id: Optional[list[str]] = Field(default=None, alias="sampleId")
    values: Union[list[float], list[list[float]]]
    high_values: Optional[Union[list[float], list[list[float]]]] = Field(
        default=None, alias="highValues"
    )
    low_values: Optional[Union[list[float], list[list[float]]]] = Field(
        default=None, alias="lowValues"
    )

    @model_validator(mode="after")
    def generate_sample_ids(self):
        """Generates sample ids if not provided"""
        if self.sample_id is None:
            self.sample_id = [str(i) for i in range(len(self.values))]
        return self


class PredictRequest(BaseModel):
    """Pydantic class representing the expected Predict Request"""

    instances: list[list[Optional[Instance]]]
    metadata: Optional[dict] = None
    config: Optional[dict] = None


class PredictResponse(BaseModel):
    """Pydantic class representing the expected Predict Response"""

    instances: list[list[Optional[Instance]]]


def urljoin(*args) -> str:
    """join url elements together into one url"""
    elements = [
        f"{arg}/" if arg[-1] != "/" and i != len(args) - 1 else arg
        for i, arg in enumerate(args)
    ]
    return reduce(urlparse.urljoin, elements)


def validate_list_elements(arr: list) -> bool:
    """Validates if an array contains non float values"""
    return any(
        not isinstance(item, (float, int))
        or item is None
        or np.isnan(item)
        or np.isinf(item)
        for item in arr
    )


def get_id_list(json_list: list[dict]) -> list[str]:
    """Extracts the id from a list of dictionaries"""
    return [item["id"] for item in json_list]


def is_date_in_format(date_string, date_format):
    """Checks if a date string is in a specific format"""
    try:
        datetime.strptime(date_string, date_format)
        return True
    except ValueError:
        return False



"""This module contains utility functions for data validation and formatting in the SDK
"""

import math

# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Protocol, Union

import numpy as np

from dhl_sdk._spectra_utils import (
    SpectraData,
    SpectraModel,
    _convert_to_request,
    _validate_spectra_format,
)
from dhl_sdk._utils import Predictions, PredictResponse, validate_list_elements
from dhl_sdk.exceptions import (
    InvalidInputsException,
    InvalidSpectraException,
    InvalidStepsException,
    InvalidTimestampsException,
)


class Group(Protocol):
    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring
    @property
    def code(self) -> str:
        ...


class Variable(Protocol):
    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring
    @property
    def group(self) -> Group:
        ...

    @property
    def code(self) -> str:
        ...

    @property
    def id(self) -> str:
        ...

    def matches_key(self, key: str) -> bool:
        ...


class Dataset(Protocol):
    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring
    @property
    def variables(self) -> list[Variable]:
        ...


class Model(Protocol):
    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring
    @property
    def dataset(self) -> Dataset:
        ...

    @property
    def model_variables(self) -> list[Variable]:
        ...


# Input Preprocessors (Validation and Formatting)
class Preprocessor(ABC):
    """Interface for preprocessing"""

    @abstractmethod
    def validate(self) -> bool:
        """Validate the input data"""

    @abstractmethod
    def format(self) -> list[dict]:
        """Format the input data to a list of JSON requests"""


@dataclass
class SpectraPreprocessor(Preprocessor):
    """Implementation of the preprocessor for Spectrum models"""

    spectra: SpectraData
    inputs: Optional[dict]
    model: SpectraModel

    def validate(self) -> bool:
        """
        This function validates and makes the necessary formatting of the spectra
        and inputs used for prediction. It performs the following validations:

        - Validates if spectra are empty
        - Validates if the number of wavelengths in the spectra match the
            number of wavelengths in the model
        - Validates if the spectra contains None (missing) values
        - Validates if the number of inputs matches model inputs
        - Validates if the number of values in each input matches the number of spectra

        Parameters
        ----------
        spectra : Union[list[list[float]], np.ndarray]
            The spectra to be validated
        model : Model
            The model to use for prediction.
        inputs : dict, optional
            A dictionary of input variables and their values, by default None

        Returns
        -------
        (list[list[float]], dict)
            If all validations pass, return the formatted spectra and inputs
            as a tuple.

        Raises
        ------
        InvalidSpectraException
            Exception raised when Spectra are not valid for prediction.
        InvalidInputsException
            Exception raised when Inputs are not valid for prediction.
        """

        # Validate if empty spectra
        n_spectra = len(self.spectra)
        if n_spectra < 1:
            raise InvalidSpectraException("Empty spectra provided")

        self.spectra = _validate_spectra_format(self.spectra)

        # Validate number of wavelengths in spectra
        for i, spectrum in enumerate(self.spectra):
            if len(spectrum) != self.model.spectra_size:
                raise InvalidSpectraException(
                    (
                        f"Invalid Spectra: The Number of Wavelengths does not"
                        f"match training data for spectrum number: {i+1}. "
                        f"Expected: {self.model.spectra_size}, Got: {len(spectrum)}"
                    )
                )
            if validate_list_elements(spectrum):
                raise InvalidSpectraException(
                    (
                        f"Invalid Spectra: The Spectra contains not "
                        f"valid values for spectrum number: {i+1}"
                    )
                )

        if self.inputs is None:
            if len(self.model.inputs) > 0:
                raise InvalidInputsException(
                    "The model requires inputs, but none were provided."
                )
            return True

        if len(self.model.inputs) == 0:
            raise InvalidInputsException(
                "The model does not require inputs, but some were provided."
            )

        # validate inputs with spectra inputs (number of lines)
        for key, value in self.inputs.items():
            if len(value) != n_spectra:
                raise InvalidInputsException(
                    f"The Number of values does not match the number of spectra for input: {key}"
                )
            if validate_list_elements(value):
                raise InvalidInputsException(
                    f"Invalid Inputs: The Inputs contain non-valid values for input: {key}"
                )

        return True

    def format(self) -> list[dict]:
        """
        Format the inputs for a given model.

        Returns
        -------
        list[dict]
            A list of JSON requests, where each request contains a list of instances.

        Raises
        ------
        InvalidInputsException
            If no matching input is found for a given key.
        """

        if self.inputs is None:
            return _convert_to_request(self.spectra, model=self.model)

        model_variables = self.model.dataset.variables

        input_variables = [
            variable for variable in model_variables if variable.id in self.model.inputs
        ]

        # validate inputs codes and format for ids
        formatted_inputs = {}
        for key, value in self.inputs.items():
            for variable in input_variables:
                if variable.matches_key(key):
                    formatted_inputs[variable.id] = value
                    break
            else:
                correct_inputs = [print(variable) for variable in input_variables]
                raise InvalidInputsException(
                    (
                        f"No matching Input found for key: {key}. "
                        f"Please select one of the following as inputs: {*correct_inputs,}"
                    )
                )

        return _convert_to_request(
            self.spectra, model=self.model, inputs=formatted_inputs
        )


@dataclass
class CultivationPropagationPreprocessor(Preprocessor):
    """Implementation of the preprocessing strategy for cultivation models"""

    timestamps: Union[list[Union[int, float]], np.ndarray]
    timestamps_unit: str
    inputs: dict[str, list]
    model: Model

    def validate(self) -> bool:
        """
        This method validates the timestamps and inputs used for prediction.
        It performs the following validations:

        - Validates the structure and values of the timestamps
            - Valid timestamps should be a list of integers, have a length of at least 2,
            contain valid numeric values,be in ascending order, have positive values and be unique.
        - Validates the structure and values of the inputs
            - The inputs should be organized with variable codes as
                keys and lists of input values as values.
        - Validates the inputs based on the model variables
            - Garantees that all the mandatory model variables are present in the inputs
            - Garantees that the inputs that need to be complete for the recipe
                have the same length as the timestamps
            - Garantees that the initial condition variables only have one value

        Returns
        -------
        bool
            True if all validations pass

        Raises
        ------
        InvalidTimestampsException
            Exception raised when Timestamps are not valid for prediction.
        InvalidInputsException
            Exception raised when Inputs are not valid for prediction.
        """

        if isinstance(self.timestamps, np.ndarray):
            self.timestamps = self.timestamps.tolist()

        # Validate timestamps
        self.timestamps = _validate_upstream_timestamps(
            timestamps=self.timestamps, timestamps_unit=self.timestamps_unit
        )

        # Validate inputs
        _validate_upstream_inputs(inputs=self.inputs)

        # validate inputs with model variables
        _validate_propagation_with_variables(self.timestamps, self.inputs, self.model)
        return True

    def format(self) -> list[dict]:
        """
        This method formats the timestamps and inputs to be in the format required by the API.:

        Returns
        -------
        list[dict]
            list of dictionaries with instances for prediction
        """

        input_variables = self.model.dataset.variables
        model_variables = self.model.model_variables

        instances = [[]]

        formatted_inputs = {}

        # order the dict according to Variables and insert timestamps
        for key, value in self.inputs.copy().items():
            for variable in model_variables:
                if variable.matches_key(key):
                    formatted_inputs[variable.id] = {}
                    formatted_inputs[variable.id]["values"] = value
                    formatted_inputs[variable.id]["timestamps"] = self.timestamps[
                        : len(value)
                    ]
                    break

        for variable in input_variables:
            if variable.id in formatted_inputs:
                instances[0].append(formatted_inputs[variable.id])
            else:
                instances[0].append(None)

        json_data = {
            "instances": instances,
            "config": {"startingIndex": 0},
        }

        return [json_data]


@dataclass
class CultivationHistoricalPreprocessor(Preprocessor):
    """Implementation of the preprocessing strategy for cultivation models"""

    timestamps: Union[list[Union[int, float]], np.ndarray]
    timestamps_unit: str
    steps: list[Optional[int]]
    inputs: dict[str, list]
    model: Model

    def validate(self) -> bool:
        """
        This method validates the timestamps and inputs used for prediction.
        It performs the following validations:

        - Validates the structure and values of the timestamps
            - Valid timestamps should be a list of integers, have a length of at least 2,
            contain valid numeric values, in ascending order, have positive values and be unique.
        - Validates the structure and values of steps
            - Valid steps should be a list of integers, contain valid numeric values, in ascending \
                  order, have positive values and be unique.
        - Validates the structure and values of the inputs
            - The inputs should be organized with variable codes as \
                keys and lists of input values as values.

        Returns
        -------
        bool
            True if all validations pass

        Raises
        ------
        InvalidTimestampsException
            Exception raised when Spectra is not valid for prediction.
        InvalidInputsException
            Exception raised when Inputs is not valid for prediction.
        """

        if isinstance(self.timestamps, np.ndarray):
            self.timestamps = self.timestamps.tolist()

        # Validate timestamps
        self.timestamps = _validate_upstream_timestamps(
            timestamps=self.timestamps, timestamps_unit=self.timestamps_unit
        )

        # Validate steps
        _validate_historical_steps(steps=self.steps, timestamps=self.timestamps)

        # Validate inputs
        _validate_upstream_inputs(inputs=self.inputs)

        # # validate inputs with model variables
        _validate_historical_with_variables(self.inputs, self.model)

        return True

    def format(self) -> list[dict]:
        """
        This method formats the timestamps and inputs to be in the format required by the API.

        Returns
        -------
        list[dict]
            list of dictionaries with instances for prediction
        """

        input_variables = self.model.dataset.variables
        model_variables = self.model.model_variables

        instances = [[]]

        formatted_inputs = {}

        # order the dict according to Variables and insert timestamps and steps
        for key, value in self.inputs.copy().items():
            for variable in model_variables:
                if variable.matches_key(key):
                    formatted_inputs[variable.id] = {}
                    formatted_inputs[variable.id]["values"] = value
                    formatted_inputs[variable.id]["timestamps"] = self.timestamps[
                        : len(value)
                    ]
                    formatted_inputs[variable.id]["steps"] = self.steps[: len(value)]
                    break

        for variable in input_variables:
            if variable.id in formatted_inputs:
                instances[0].append(formatted_inputs[variable.id])
            else:
                instances[0].append(None)

        json_data = {"instances": instances, "config": {"startingIndex": 0}}

        return [json_data]


def _validate_upstream_timestamps(
    timestamps: list[Union[int, float]], timestamps_unit: str
) -> list[Union[int, float]]:
    """Validate the timestamps for upstream prediction.

    This function performs a series of validations on the provided timestamps to ensure they meet
    the requirements for upstream prediction tasks.

    Parameters
    ----------
    timestamps : list[Union[int, float]]
        List of timestamps in the specified unit of time
    timestamps_unit : str
        Unit of the timestamps. Can be one of the following:
        - 's', 'sec', 'secs', 'seconds'
        - 'm', 'min', 'mins', 'minutes'
        - 'h', 'hour', 'hours'
        - 'd', 'day', 'days'

    Returns
    -------
    list[Union[int, float]]
        List of timestamps in seconds

    Raises
    ------
    InvalidTimestampsException
        If timestamps do not meet the specified criteria.

    """
    # Validate types
    if not isinstance(timestamps, list):
        raise InvalidTimestampsException("Timestamps must be a list of numbers")

    # Validate if length of timestamps
    if len(timestamps) <= 1:
        raise InvalidTimestampsException(
            "Timestamps must be a list of at least 2 values"
        )

    # Validate if timestamps are valid numeric values
    if not all(
        isinstance(value, (int, float)) and math.isfinite(value) for value in timestamps
    ):
        raise InvalidTimestampsException(
            "All values of timestamps must be valid numeric values"
        )

    # Validate if timestamps are in ascending order and unique
    if any(a >= b for a, b in zip(timestamps, timestamps[1:])):
        raise InvalidTimestampsException("Timestamps must be in ascending order")

    # Validate if timestamps are positive (Since they are ordered, just check the first one)
    if timestamps[0] < 0:
        raise InvalidTimestampsException("Timestamps must be positive")

    # Convert timestamps to seconds
    if timestamps_unit.lower() in ("s", "sec", "secs", "seconds"):
        return timestamps
    if timestamps_unit.lower() in ("m", "min", "mins", "minutes"):
        factor = 60
        return [timestamp * factor for timestamp in timestamps]
    if timestamps_unit.lower() in ("h", "hour", "hours"):
        factor = 60 * 60
        return [timestamp * factor for timestamp in timestamps]
    if timestamps_unit.lower() in ("d", "day", "days"):
        factor = 60 * 60 * 24
        return [timestamp * factor for timestamp in timestamps]

    raise InvalidTimestampsException(
        f"Invalid timestamps unit '{timestamps_unit}' found."
    )


def _validate_historical_steps(
    steps: list[Optional[int]], timestamps: list[Union[int, float]]
) -> None:
    """Validate the steps for historical prediction.

    This function performs a series of validations on the provided timestamps to ensure they meet
    the requirements for upstream prediction tasks.

    Parameters
    ----------
    steps : list[Optional[int]]
        List of steps, if None it means that a timestamp is not annotated with a step
    timestamps : list[Union[int, float]]
        List of timestamps corresponding to the steps

    Raises
    ------
    InvalidStepsException
        If steps do not meet the specified criteria.


    """
    # Validate types
    if not isinstance(steps, list):
        raise InvalidStepsException("Steps must be a list of numbers")

    # Validate if length of steps matches timestamps
    if len(steps) != len(timestamps):
        raise InvalidStepsException("Steps must have the same length as timestamps")

    # Validate if steps are in ascending order and unique
    filtered_steps = [step for step in steps if step is not None]

    if any(a >= b for a, b in zip(filtered_steps, filtered_steps[1:])):
        raise InvalidStepsException("Steps must be in ascending order")

    # Validate if steps start at 0
    if filtered_steps[0] != 0:
        raise InvalidStepsException("Steps must start at 0")


def _validate_upstream_inputs(inputs: dict[str, list]) -> None:
    """
    Validate the inputs for upstream prediction.
    The function checks if the inputs are a dictionary and if all values are lists.

    Parameters
    ----------
    inputs : dict[str, list]
        A dictionary, or list of dictionaries, where keys are variable codes,
        and values are lists of inputs.

    Raises
    ------
    InvalidInputsException
        If inputs do not meet the specified criteria.
    """

    if inputs is None:
        raise InvalidInputsException(
            "No Inputs provided. Please provide a dictionary of inputs"
        )

    # Validate types
    if not isinstance(inputs, dict):
        raise InvalidInputsException(
            "Inputs must be a dictionary of lists, with the variable code as key"
        )

    # Validate if inputs are lists
    if not all(isinstance(value, list) for value in inputs.values()):
        raise InvalidInputsException("All input values must be lists")


def _validate_propagation_with_variables(
    timestamps: list[Union[int, float]], inputs: dict[str, list], model: Model
) -> None:
    """
    Validate the inputs based on the provided timestamps and model.

    Parameters
    ----------
    timestamps : list[int]
        List of timestamps
    inputs : dict[str, list] | list[dict[str, list]]
        A dictionary where keys are variable codes, and values are lists of inputs.
    model : Model
        Model used for prediction

    Raises
    ------
    InvalidInputsException
        If inputs do not meet the specified criteria.
    """

    model_variables = model.model_variables

    for variable in model_variables:
        if variable.code in inputs:
            if groupcode_is_timedependent(variable.group.code):
                # Validate if inputs and timestamps are the same length
                if len(inputs[variable.code]) != len(timestamps):
                    raise InvalidInputsException(
                        (
                            f"The recipe requires {variable.code} to be complete,"
                            f"so it must have the same length as timestamps"
                        )
                    )

            else:
                # validate if non time dependent inputs have length of 1 (only initial values)
                if len(inputs[variable.code]) != 1:
                    raise InvalidInputsException(
                        (
                            f"Input {variable.code} only requires initial "
                            f"values, so it must have a length of 1"
                        )
                    )

            # Validate if numeric inputs are valid numeric values
            if groupcode_is_numeric(variable.group.code):
                if not all(
                    isinstance(value, (int, float)) and math.isfinite(value)
                    for value in inputs[variable.code]
                ):
                    raise InvalidInputsException(
                        f"All values of input {variable.code} must be valid numeric values"
                    )

        else:
            # validate if missing value is an Y Variable
            if not groupcode_is_output(variable.group.code):
                raise InvalidInputsException(
                    (
                        f"Input {variable.code} is a {variable.group.code} "
                        f"Variable, so it must be provided"
                    )
                )


def _validate_historical_with_variables(
    inputs: dict[str, list],
    model: Model,
) -> None:
    """
    Validate the inputs based on the provided timestamps and model.

    Parameters
    ----------
    timestamps : list[int]
        List of timestamps
    inputs : dict[str, list]
        A dictionary where keys are variable codes, and values are lists of inputs.
    model : Model
        Model used for prediction

    Raises
    ------
    InvalidInputsException
        If inputs do not meet the specified criteria.
    """

    model_variables = model.model_variables

    for variable in model_variables:
        if variable.code in inputs:
            if groupcode_is_numeric(variable.group.code):
                # Validate if inputs are valid numeric values
                if not all(
                    isinstance(value, (int, float)) for value in inputs[variable.code]
                ):
                    raise InvalidInputsException(
                        f"All values of input {variable.code} must be valid numeric values"
                    )

        else:
            # validate if missing value is an Y Variable
            if not groupcode_is_output(variable.group.code):
                raise InvalidInputsException(
                    (
                        f"Input {variable.code} is a {variable.group.code} "
                        f"Variable, so it must be provided"
                    )
                )


def format_predictions(predictions: list[PredictResponse], model: Model) -> Predictions:
    """Format a list of predictions into a dictionary.

    Parameters
    ----------
    predictions : List[PredictResponse]
        list of predictions from the API.
    model : Model
        Model used for prediction

    Returns
    -------
    Dictionary with predictions where:
        key: variable id
        value: dictionary with keys "values" and, if applicable, "upperBound" and "lowerBound"

    """

    variables = [var.code for var in model.dataset.variables]

    dic = {}

    for pred in predictions:
        for i, instance in enumerate(pred.instances[0]):
            if instance is not None:
                if variables[i] not in dic:
                    dic[variables[i]] = {}

                    dic[variables[i]]["values"] = instance.values.copy()

                    if instance.high_values is not None:
                        dic[variables[i]]["upperBound"] = instance.high_values.copy()
                    if instance.low_values is not None:
                        dic[variables[i]]["lowerBound"] = instance.low_values.copy()
                else:
                    dic[variables[i]]["values"].extend(instance.values)

                    if instance.high_values is not None:
                        dic[variables[i]]["upperBound"].extend(instance.high_values)
                    if instance.low_values is not None:
                        dic[variables[i]]["lowerBound"].extend(instance.low_values)

    return dic


def variant_is_numeric(variant: str) -> bool:
    """Check if the variant is numeric"""
    return variant in ["flow", "numeric"]


def groupcode_is_numeric(code: str):
    """Check if the group is numeric"""
    return code in [
        "Flows",
        "FeedConc",
        "Inducers",
        "W",
        "X",
    ]


def groupcode_is_output(code: str):
    """Check if the group is an output / CQA"""
    return code == "Y"


def groupcode_is_timedependent(code: str):
    """Check if the group is time dependent for recipe only"""
    return code in ["Flows", "W", "Inducers"]



"""
Exceptions Module

This module defines custom exception classes for use throughout the project.

Exceptions:
    - InvalidSpectraException: Exception raised when spectra used for prediction 
      are not valid
    - InvalidInputsException: Exception raised when Inputs used for prediction
      are not valid
    - ModelPredictionException: Exception raised when Model Prediction fails
    - InvalidTimestampsException: Exception raised when timestamps used for 
      prediction are not valid
    - InvalidStepsException: Exception raised when steps used for historical 
      model prediction are not valid
    - PredictionRequestException: Exception raised when prediction request fails
    - ImportValidationException: Exception raised when import validation fails
    - NewEntityException: Exception raised when creating new entity fails
"""


class InvalidSpectraException(Exception):
    """Exception raised when spectra used for prediction are not valid"""

    def __init__(self, message="The Spectra used for prediction are not valid."):
        self.message = message
        super().__init__(self.message)


class InvalidInputsException(Exception):
    """Exception raised when Inputs used for prediction are not valid"""

    def __init__(self, message="The Inputs used for prediction are not valid."):
        self.message = message
        super().__init__(self.message)


class ModelPredictionException(Exception):
    """Exception raised when Model Prediction fails"""

    def __init__(self, message="Model is not valid for prediction."):
        self.message = message
        super().__init__(self.message)


class InvalidTimestampsException(Exception):
    """Exception raised when timestamps used for prediction are not valid"""

    def __init__(self, message="The timestamps used for prediction are not valid."):
        self.message = message
        super().__init__(self.message)


class InvalidStepsException(Exception):
    """Exception raised when steps used for historical model prediction are not valid"""

    def __init__(self, message="The steps used for the prediction are not valid."):
        self.message = message
        super().__init__(self.message)


class InvalidStartingIndexException(Exception):
    """Exception raised when starting index given is not valid"""

    def __init__(self, message="The starting index is not valid."):
        self.message = message
        super().__init__(self.message)


class PredictionRequestException(Exception):
    """Exception raised when prediction request fails"""

    def __init__(self, message="Prediction request failed."):
        self.message = message
        super().__init__(self.message)


class ImportValidationException(Exception):
    """Exception raised when import validation fails"""

    def __init__(self, message="Import validation failed."):
        self.message = message
        super().__init__(self.message)


class NewEntityException(Exception):
    """Exception raised when creating new entity fails"""

    def __init__(self, message="Creating new entity failed."):
        self.message = message
        super().__init__(self.message)



"""API Entities Module

This module provides a comprehensive set of Pydantic models that represent
multiple entities that can be downloaded or uploaded to the DataBase through
the API . 

Classes:
    - Experiment: Represents a structure for experiments
    - Recipe: Represents a structure for recipes
    - Variable: Represents a structure for variables present in experiments and recipes
    - Product: Represents a structure for products present in experiments and recipes
    """

# pylint: disable=no-member, protected-access

import copy
import csv
from abc import ABC, abstractmethod
from enum import Enum
from io import StringIO
from typing import Literal, Optional, Union

from pydantic import BaseModel, Field, PrivateAttr, model_validator
from requests import Response

from dhl_sdk._utils import (
    EXPERIMENTS_URL,
    FILES_URL,
    PRODUCTS_URL,
    RECIPES_URL,
    VARIABLES_URL,
    VariableGroupCodes,
)
from dhl_sdk.crud import Client, CRUDClient, DataBaseClient
from dhl_sdk.exceptions import ImportValidationException, NewEntityException
from dhl_sdk.importers import RunFileImporter, SpectraFileImporter
from dhl_sdk.validators import (
    AbstractFileValidator,
    AbstractValidator,
    ExperimentFileValidator,
    ExperimentValidator,
    ProductValidator,
    RecipeFileValidator,
    SpectraFileValidator,
    VariableValidator,
)


class DataBaseEntity(ABC):
    """Abstract class for DataBase Entities

    Database Entities are the entities that can be pushed and pulled
    directly from the database.
    """

    @abstractmethod
    def validate_import(self, client: Client) -> bool:
        """Validate the entity for importing"""

    @abstractmethod
    def create_request_body(self) -> dict:
        """Create request body for creating the entity"""

    @staticmethod
    @abstractmethod
    def requests(client: Client) -> CRUDClient:
        """Return a CRUDClient for the entity"""


class VariableNumeric(BaseModel):
    """Model for Variables that can be caracterized by a numeric value.

    Attributes:
        default: Default value
        minimum: Minimum value
        maximum: Maximum value
        interpolation_method: Interpolation method for the variable

    """

    default: Optional[float] = Field(default=None, alias="default")
    minimum: Optional[float] = Field(default=None, alias="min")
    maximum: Optional[float] = Field(default=None, alias="max")
    interpolation_method: Optional[Literal["linear", "discrete"]] = Field(
        default=None, alias="interpolation"
    )

    @property
    def variant_string(self) -> str:
        """get variant type as string"""
        return "numeric"


class VariableCategorical(BaseModel):
    """Model for Categorical Variables

    Attributes:
        default: Default value
        strict: Strict mode (whether to allow only values from the categories)
        categories: Categories for the variable
    """

    default: Optional[str] = Field(default=None, alias="default")
    strict: Optional[bool] = Field(default=False, alias="strict")
    categories: Optional[list[str]] = Field(default=None, alias="values")

    @property
    def variant_string(self) -> str:
        """get variant type as string"""
        return "categorical"


class VariableLogical(BaseModel):
    """Model for Logical Variables

    Attributes:
        default: Default value
    """

    default: Optional[bool] = Field(default=None, alias="default")

    @property
    def variant_string(self) -> str:
        """get variant type as string"""
        return "logical"


class FlowVariableReference(BaseModel):
    """Model for Flow Variable References with only the ID, used to serialize the data

    Attributes:
        measurement_id: ID of the measurement variable (X Variable)
        concentration_id: ID of concentration variable (Feed Concentration)
        fraction_id: ID of fraction
    """

    measurement_id: str = Field(alias="measurementId")
    concentration_id: Optional[str] = Field(default=None, alias="concentrationId")
    fraction_id: Optional[str] = Field(default=None, alias="fractionId")


class VariableFlow(BaseModel):
    """Model for Flow Variables

    Attributes:
        flow_type: Type of flow variable - can be "bolus", "conti",
            "mbolus", "mconti", "sampling", "bleed", "perfusion"
        step_size: Step size for the flow variable
        volume_id: Volume ID for the flow variable
        references: References for the flow variable
    """

    flow_type: Literal[
        "bolus", "conti", "mbolus", "mconti", "sampling", "bleed", "perfusion"
    ] = Field(default="bolus", alias="type")
    step_size: Optional[int] = Field(default=None, alias="stepSize")
    volume_id: Optional[str] = Field(default=None, alias="volumeId")
    references: list[FlowVariableReference] = Field(alias="references")

    @staticmethod
    def new(
        flow_type: Literal[
            "bolus", "conti", "mbolus", "mconti", "sampling", "bleed", "perfusion"
        ],
        variable_references: list[FlowVariableReference],
        step_size: Optional[int] = None,
        volume_variable_id: Optional[str] = None,
    ) -> "VariableFlow":
        """Create a new Flow Variable from user input

        Parameters
        ----------
        flow_type : Literal["bolus", "conti", "mbolus", "mconti", "sampling", "bleed", "perfusion"]
            Type of flow variable
        variable_references : list[FlowVariableReference]
            References for the flow variable. It must contain at least one reference.
            Each Reference must be of type FlowVariableReference (or a dictionary
            with the same keys)
            FlowVariableReference contains the following keys:
                - measurement_id: ID of the measurement variable (X Variable)
                - concentration_id: ID of concentration variable (Feed Concentration)
        step_size : Optional[int]
            Step size for the flow variable in seconds. This is required for "conti", "mconti",
            "bleed", "perfusion" types
        volume_variable_id : Optional[str]
            Variable ID of the Initial Bioreactor Volume
        """

        if flow_type in ["conti", "mconti", "bleed", "perfusion"]:
            if not step_size:
                raise NewEntityException(
                    "Step size must be provided for 'conti', 'mconti', 'bleed', 'perfusion' types"
                )

        return VariableFlow(
            type=flow_type,
            stepSize=step_size,
            volumeId=volume_variable_id,
            references=variable_references,
        )

    @property
    def variant_string(self) -> str:
        """get variant type as string"""
        return "flow"


class VariableSpectrumXAxis(BaseModel):
    """Pydantic model for variant spectrum x axis"""

    dimension: Optional[int] = Field(default=None, alias="dimension")
    unit: Optional[str] = Field(default=None, alias="unit")
    min: Optional[float] = Field(default=None, alias="min")
    max: Optional[float] = Field(default=None, alias="max")


class VariableSpectrumYAxis(BaseModel):
    """Pydantic model for variant spectrum y axis"""

    label: Optional[str] = Field(default=None, alias="label")


class VariableSpectrum(BaseModel):
    """Pydantic model for variant details spectrum

    Attributes:
        x_axis: X axis description (dimension, unit, min, max)
        y_axis: Y axis description (label)
    """

    x_axis: Optional[VariableSpectrumXAxis] = Field(default=None, alias="xAxis")
    y_axis: Optional[VariableSpectrumYAxis] = Field(default=None, alias="yAxis")

    @property
    def variant_string(self) -> str:
        """get variant type as string"""
        return "spectrum"


VariantDetails = Union[
    VariableNumeric,
    VariableCategorical,
    VariableLogical,
    VariableFlow,
    VariableSpectrum,
]


class Group(BaseModel):
    """Model for Variable Group"""

    id: Optional[str] = Field(default=None, alias="id")
    code: Optional[str] = Field(default=None, alias="code")
    name: Optional[str] = Field(default=None, alias="name")
    aggregation: Optional[str] = Field(default=None, alias="aggregation")

    def validate_group(self, group_codes: dict[str, tuple[str, str]]) -> None:
        """Validate the group"""
        if self.name not in group_codes:
            raise ImportValidationException(
                (
                    f"Variable Group must be one of: {list(group_codes.keys())}."
                    " instead, it got '{self.name}'"
                )
            )

        self.id, self.code = group_codes[self.name]


class Variant(str, Enum):
    """Enum for variants"""

    FLOW = "flow"
    NUMERIC = "numeric"
    CATEGORICAL = "categorical"
    LOGICAL = "logical"
    SPECTRUM = "spectrum"

    def get_variant_details(self, **kwargs) -> VariantDetails:
        """get variant details"""
        if self == Variant.FLOW:
            return VariableFlow(**kwargs)
        if self == Variant.NUMERIC:
            return VariableNumeric(**kwargs)
        if self == Variant.CATEGORICAL:
            return VariableCategorical(**kwargs)
        if self == Variant.LOGICAL:
            return VariableLogical(**kwargs)
        if self == Variant.SPECTRUM:
            return VariableSpectrum(**kwargs)

        raise ValueError(f"Variant {self} not supported")


class Variable(BaseModel, DataBaseEntity):
    """Model Variable"""

    id: Optional[str] = Field(default=None, alias="id")
    name: str = Field(alias="name")
    code: str = Field(alias="code")
    description: Optional[str] = Field(default="", alias="description")
    measurement_unit: Optional[str] = Field(default=None, alias="measurementUnit")
    group: Optional[Group] = Field(default=None, alias="group")
    variant: Variant = Field(alias="variant")
    variant_details: Optional[VariantDetails] = None
    size: Optional[int] = None

    _validator: AbstractValidator = PrivateAttr(VariableValidator())

    @model_validator(mode="before")  # type: ignore
    @classmethod
    def _validate_model_struct(cls, data) -> dict:
        """Validate the structure of the variable"""

        variant = Variant(data["variant"])

        # check if variant details field is in input data
        if "variant_details" not in data:
            variant_details = data.get(variant.value, {})
            data["variant_details"] = variant.get_variant_details(**variant_details)
        else:
            try:
                if data["variant_details"].variant_string != variant.value:
                    raise ValueError(
                        (
                            f"Variant details not valid for {variant} variant."
                            f"Found: {data['variant_details'].variant_string}"
                        )
                    )
            except AttributeError as asc:
                raise ValueError(
                    f"Variant details not valid for {variant} variant"
                ) from asc

        if variant == Variant.SPECTRUM:
            try:
                size = data["spectrum"]["xAxis"]["dimension"]
            except KeyError as err:
                raise KeyError(
                    "The spectrum variable does not have a valid structure"
                ) from err

            data["size"] = size

        return data

    def matches_key(self, key) -> bool:
        """Find the id of the variable"""
        if self.id == key or (self.code is not None and self.code == key):
            return True
        return False

    def __str__(self) -> str:
        """Print only the variable's ID and Code"""
        return f"Name: {self.name} ,  Code: {self.code}"

    def validate_import(self, client: Client) -> bool:
        """Validate the variable for uploading"""

        # validate and update group
        variable_group_codes = VariableGroupCodes(client).get_variable_group_codes()

        if self.group:
            self.group.validate_group(variable_group_codes)
        else:
            raise ImportValidationException(
                "Variable group is not present, please provide a valid group"
            )

        # validate entity
        return self._validator.validate(entity=self, client=client)

    def create_request_body(self) -> dict:
        """Create request body for creating a variable"""

        model_dict = self.model_dump(
            by_alias=True,
            include={
                "code": True,
                "name": True,
                "description": True,
                "measurement_unit": True,
                "group": {"id"},
            },
        )

        model_dict["variant"] = self.variant.value

        if self.variant_details:
            model_dict[self.variant.value] = self.variant_details.model_dump(
                by_alias=True, exclude_none=True
            )
        else:
            model_dict[self.variant.value] = {}

        return model_dict

    def is_imported(self, client: Client) -> bool:
        """Check if the variable is already imported"""

        if not self.id:
            return False

        response = client.get(f"{VARIABLES_URL}/{self.id}")
        return response.status_code == 200

    @staticmethod
    def get_valid_variable_groups(client: DataBaseClient) -> list[str]:
        """Get the valid variable groups from the database

        It is a helper function to assist with the creation of new variables.
        """
        group_codes = VariableGroupCodes(client._client).get_variable_group_codes()
        return list(group_codes.keys())

    @staticmethod
    def new(
        code: str,
        name: str,
        measurement_unit: str,
        variable_group: str,
        variable_type: Union[
            VariableNumeric,
            VariableCategorical,
            VariableLogical,
            VariableFlow,
            VariableSpectrum,
        ],
        description: Optional[str] = "",
    ) -> "Variable":
        """Create a new variable from user input

        Parameters
        ----------
        code : str
            Code of the variable (must be unique and from 1 to 5 characters long)
        name : str
            Name of the variable
        measurement_unit : str
            Measurement unit of the variable
        variable_group : str
            Group of the variable
        variable_type : Union[VariableNumeric,
                              VariableCategorical,
                              VariableLogical,
                              VariableFlow,
                              VariableSpectrum]
            Type of the variable
        description : Optional[str]
            Description of the variable

        Returns
        -------
        Variable
            A new variable with the given parameters
        """

        code = code.strip()
        name = name.strip()
        measurement_unit = measurement_unit.strip()
        variable_group = variable_group.strip()

        if code == "":
            raise NewEntityException("Variable code can not be empty")

        if name == "":
            raise NewEntityException("Variable name can not be empty")

        if measurement_unit == "":
            raise NewEntityException("Measurement unit can not be empty")

        if variable_group == "":
            raise NewEntityException("Variable group can not be empty")

        var_dict = {
            "code": code,
            "name": name,
            "description": description if description is not None else "",
            "variant": variable_type.variant_string,
            "measurementUnit": measurement_unit,
            "group": {
                "name": variable_group,
            },
            "variant_details": variable_type,
        }

        return Variable(**var_dict)

    @staticmethod
    def requests(client: Client) -> CRUDClient["Variable"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["Variable"](client, VARIABLES_URL, Variable)


class Product(BaseModel, DataBaseEntity):
    """Pydantic model for Product"""

    id: Optional[str] = Field(default=None, alias="id")
    code: str = Field(alias="code")
    name: str = Field(alias="name")
    description: Optional[str] = Field(default="", alias="description")

    _validator: AbstractValidator = PrivateAttr(ProductValidator())

    def validate_import(self, client: Client) -> bool:
        """Validate if the product can be imported"""
        return self._validator.validate(entity=self, client=client)

    def create_request_body(self) -> dict:
        """Create request body for creating a product"""
        return self.model_dump(exclude_none=True, by_alias=True, exclude={"id"})

    @staticmethod
    def new(
        code: str,
        name: str,
        description: Optional[str] = "",
    ) -> "Product":
        """Create a new Product from user input

        Parameters
        ----------
        code : str
            Code of the product (must be unique and from 1 to 5 characters long)
        name : str
            Name of the product
        description : Optional[str]
            Description of the product

        Returns
        -------
        Product
            A new product with the given parameters

        """

        code = code.strip()
        name = name.strip()

        if code == "":
            raise NewEntityException("Product code cannot be empty")

        if name == "":
            raise NewEntityException("Product name cannot be empty")

        if len(code) > 5:
            raise NewEntityException("Product code must be from 1 to 5 characters long")

        return Product(code=code, name=name, description=description)

    @staticmethod
    def requests(client: Client) -> CRUDClient["Product"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["Product"](client, PRODUCTS_URL, Product)


class File(BaseModel):
    """Pydantic model for File"""

    id: Optional[str] = Field(default=None, alias="id")
    name: str = Field(alias="name")
    description: str = Field(alias="description")
    type: str = Field(alias="type")
    variant: Literal["run", "samples"] = Field(default="run", alias="variant")

    _data: dict[str, dict[str, list]] = PrivateAttr()
    _validator: AbstractFileValidator = PrivateAttr(ExperimentFileValidator())

    def __init__(self, **data):
        super().__init__(**data)
        self._data = data["data"]
        self._validator = data["validator"]

    def validate_import(
        self, variables: list[Variable], variant_details: Optional[dict] = None
    ) -> bool:
        """Validate if the file can be imported"""
        if self._validator.validate(
            variables=variables,
            data=self._data,
            variant=self.variant,
            variant_details=variant_details,
        ):
            self._data = self._validator.format_data(
                variables=variables, data=self._data
            )
            return True
        else:
            return False

    def create_request_body(self) -> dict:
        """Create request body for creating a file"""

        file_dict = self.model_dump(
            exclude_none=True,
            by_alias=True,
            include={"name": True, "description": True},
        )

        if self.type == "run":
            suffix = "Data"
        elif self.type == "spectra":
            suffix = "Spectra"

        if self.variant == "run":
            file_dict["type"] = f"run{suffix}"
        elif self.variant == "samples":
            file_dict["type"] = f"sample{suffix}"

        return file_dict

    def create_file(self, client: Client) -> Optional[Union[str, tuple[str, str]]]:
        """Create a file in the database"""

        if self.type == "run":
            importer = RunFileImporter(client)
        elif self.type == "spectra":
            importer = SpectraFileImporter(client)
        else:
            raise ValueError("File type must be either 'run' or 'spectra'")

        file_id = importer.import_file(self)

        return file_id

    @staticmethod
    def requests(client: Client) -> CRUDClient["File"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["File"](client, FILES_URL, File)

    @staticmethod
    def download(client: Client, file_id: str) -> Response:
        # pylint: disable=missing-function-docstring
        return client.get(f"{FILES_URL}/{file_id}/data")


class Instances(BaseModel):
    """Pydantic model for Unresolved Instances"""

    column: str = Field(alias="column")
    fileId: str = Field(alias="fileId")


class Recipe(BaseModel, DataBaseEntity):
    """Pydantic model for Recipe"""

    id: Optional[str] = Field(default=None, alias="id")
    name: str = Field(alias="name")
    description: Optional[str] = Field(default=None, alias="description")
    product: Product = Field(alias="product")
    duration: Optional[int] = Field(default=None, alias="duration")
    variables: list[Variable] = Field(alias="variables")
    instances: list[Instances] = Field(alias="instances")
    file_data: Optional[File] = None

    _validator: AbstractValidator = PrivateAttr(ExperimentValidator())

    def create_request_body(self) -> dict:
        """Create request body for creating an experiment"""

        model_dict = self.model_dump(
            by_alias=True,
            include={
                "name": True,
                "description": True,
                "duration": True,
                "product": {"id"},
                "variables": {"__all__": {"id"}},
                "instances": {"__all__": {"column", "fileId"}},
            },
        )

        return model_dict

    def validate_import(self, client: Client) -> bool:
        """Validate if the recipe can be imported"""

        if not self._validator.validate(entity=self, client=client):
            return False

        if self.file_data.validate_import(self.variables):
            file_id = self.file_data.create_file(client)

        if not file_id:
            raise ImportValidationException("File data could not be imported")

        for variable in self.variables:
            self.instances.append(Instances(column=variable.code, fileId=file_id))

        return True

    @staticmethod
    def new(
        name: str,
        description: str,
        product: Product,
        variables: list[Variable],
        data: dict,
        duration: Optional[int] = None,
    ) -> "Recipe":
        """Create a new Recipe from user input

        Parameters
        ----------
        name : str
            Name of the recipe
        description : str
            Description of the recipe
        product : Product
            Product associated to the recipe (needs to be created before)
        variables : list[Variable]
            List of variables associated to the recipe (needs to be created before)
        data : dict
            Data to be uploaded to the recipe (dictionary)
        duration : Optional[int]
            Duration of the recipe

        Returns
        -------
        Recipe
            A new recipe with the given parameters

        """

        name = name.strip()
        if name == "":
            raise NewEntityException("Recipe name cannot be empty")

        # Create a new Data File
        file = File(
            name=name,
            description=description,
            type="run",
            data=copy.deepcopy(data),
            validator=RecipeFileValidator(duration=duration),
        )

        # Create a new Experiment
        recipe = Recipe(
            name=name,
            description=description,
            duration=duration,
            product=product,
            variables=variables,
            instances=[],
            file_data=file,
        )

        return recipe

    @staticmethod
    def requests(client: Client) -> CRUDClient["Recipe"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["Recipe"](client, RECIPES_URL, Recipe)


class Experiment(BaseModel, DataBaseEntity):
    """Pydantic model for Experiment"""

    id: Optional[str] = Field(default=None, alias="id")
    name: str = Field(alias="name")
    description: str = Field(alias="description")
    product: Product = Field(alias="product")
    subunit: str = Field(default="", alias="subunit")
    variables: list[Variable] = Field(alias="variables")
    instances: list[Instances] = Field(alias="instances")
    variant: Literal["run", "samples"] = Field(default="run", alias="variant")
    variant_details: Optional[dict] = None
    file_data: Optional[File] = None

    _validator: AbstractValidator = PrivateAttr(ExperimentValidator())

    def create_request_body(self) -> dict:
        """Create request body for creating an experiment"""
        model_dict = self.model_dump(
            by_alias=True,
            include={
                "name": True,
                "description": True,
                "subunit": True,
                "product": {"id"},
                "variables": {"__all__": {"id"}},
                "instances": {"__all__": {"column", "fileId"}},
                "variant": True,
            },
        )

        model_dict[self.variant] = self.variant_details

        return model_dict

    def validate_import(self, client: Client) -> bool:
        """Validate if the experiment can be imported"""

        if not self._validator.validate(entity=self, client=client):
            return False

        if self.file_data.validate_import(self.variables, self.variant_details):
            file_id = self.file_data.create_file(client)

        if not file_id:
            raise ImportValidationException("File data could not be imported")

        # if file_id is a tuple, it means that the file is a spectra file
        if isinstance(file_id, tuple):
            self.instances.append(Instances(column="1", fileId=file_id[0]))
            file_id = file_id[1]

        for variable in self.variables:
            # skip if variable is spectra
            if variable.variant == Variant.SPECTRUM:
                continue
            self.instances.append(Instances(column=variable.code, fileId=file_id))

        return True

    def get_data(self, client: DataBaseClient) -> dict:
        """Get experiment data from the API

        Parameters
        ----------
        client : DataBaseClient
            Client to use to get the data, i.e., DataHowLabClient

        Returns
        -------
        Dictionary with the data where:
            key: variable code
            value: a dictionary with the data for the variable, with "timestamps" and "values" keys

        Example:
        --------
        >>> client = DataHowLabClient()
        >>> data = experiment.get_data(client)
        >>> print(data)
        {'var1': {"timestamps": [1,2,3], "values": [1.1, 2.2, 3.3]},
        'var2': {"timestamps": [1,2,3], "values": [10.1, 12.2, 33.3]}}

        """

        cache = {}
        experiment_data = {}

        for instance, variable in zip(self.instances, self.variables):
            if instance.fileId not in cache:
                response = File.download(client._client, instance.fileId)
                cache[instance.fileId] = response
            else:
                response = cache[instance.fileId]

            if instance.column in experiment_data:
                continue

            if response.headers["content-type"] == "application/json":
                data = response.json()
                experiment_data[variable.code] = data["timeseries"][instance.column]

            elif response.headers["content-type"] == "text/csv":
                csv_data = [row for row in csv.reader(StringIO(response.text))]

                ids = [row[0] for row in csv_data]
                if self.variant == "run":  # run variant
                    spectra_ids = "timestamps"
                elif self.variant == "samples":  # samples variant
                    spectra_ids = "sampleId"

                data_rows = [row[1:] for row in csv_data]
                experiment_data["spectra"] = {
                    spectra_ids: ids,
                    "values": data_rows,
                }

        return experiment_data

    @staticmethod
    def new(
        name: str,
        description: str,
        product: Product,
        variables: list[Variable],
        data_type: Literal["run", "spectra"],
        data: dict,
        variant: Literal["run", "samples"],
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
    ) -> "Experiment":
        """Create a new Experiment from user input

        Parameters
        ----------
        name : str
            Name of the experiment
        description : str
            Description of the experiment
        product : Product
            Product associated to the experiment (needs to be created before)
        variables : list[Variable]
            List of variables associated to the experiment (needs to be created before)
        data_type : Literal["run", "spectra"]
            Type of data to be uploaded ("run" or "spectra")
        data : dict
            Data to be uploaded to the experiment (dictionary).
            The data must be a dictionary with the variable codes as keys.
            For "spectra" variant, the data must be a dictionary with the variable codes
                as keys and a list of lists of values as values.
        variant : Literal["run", "samples"]
            Variant of the experiment ("run" or "samples").
            "run" is for time series data and "samples" is for sampled data
        start_time : Optional[str]
            Start time of the experiment (only for "run" variant)
        end_time : Optional[str]
            End time of the experiment (only for "run" variant)

        Returns
        -------
        Experiment
            A new experiment with the given parameters

        """

        name = name.strip()
        if name == "":
            raise NewEntityException("Experiment name cannot be empty")

        if variant == "run" and (not start_time or not end_time):
            raise NewEntityException(
                "Start time and end time must be provided for 'run' variant"
            )

        file_validator = ExperimentFileValidator()
        if data_type == "spectra":
            file_validator = SpectraFileValidator()

        # Create a new Data File
        file = File(
            name=name,
            description=description,
            type=data_type,
            variant=variant,
            data=copy.deepcopy(data),
            validator=file_validator,
        )

        if variant == "run":
            variant_details = {"startTime": start_time, "endTime": end_time}
        else:
            variant_details = {}

        # Create a new Experiment
        experiment = Experiment(
            name=name,
            description=description,
            product=product,
            variables=variables,
            instances=[],
            file_data=file,
            variant=variant,
            variant_details=variant_details,
        )

        return experiment

    @staticmethod
    def requests(client: Client) -> CRUDClient["Experiment"]:
        # pylint: disable=missing-function-docstring
        return CRUDClient["Experiment"](client, EXPERIMENTS_URL, Experiment)



# pylint: disable=no-member
# pylint: disable=too-few-public-methods
# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring

"""API Results Handling Module

This module provides utility functions to handle and manage 
results obtained from the API. 

Classes:
    - CRUDClient: A utility class for handling CRUD requests for API entities
    - Result: A utility class for handling paginated API results
      and iterating through them.
"""

from collections import deque
from typing import Any, Generic, Optional, Protocol, TypeVar

from requests import Response


class Client(Protocol):
    def post(self, path: str, json_data: Any) -> Response:
        ...

    def put(self, path: str, data: Any, content_type: str) -> Response:
        ...

    def get(self, path: str, query_params: Optional[dict[str, str]] = None) -> Response:
        ...


class DataBaseClient(Protocol):
    """Protocol for Clients that interact with the DB, like DataHowLabClient"""

    _client: Client


T = TypeVar("T")


class Constructor(Protocol[T]):
    def __call__(self, **kwargs) -> T:
        ...


class CRUDClient(Generic[T]):
    """Utility class for handling CRUD requests for API entities"""

    def __init__(self, client: Client, base_url: str, constructor: Constructor[T]):
        self._client = client
        self._base_url = base_url
        self._constructor = constructor

    def get(self, entity_id: str) -> T:
        """Get an entity by its ID from the API"""
        response = self._client.get(f"{self._base_url}/{entity_id}")
        entity = response.json()
        entity = self._constructor(**entity, client=self._client)

        return entity

    def create(self, data: dict[str, Any]) -> T:
        """Create an entity in the API"""
        response = self._client.post(self._base_url, json_data=data)
        entity = response.json()
        entity = self._constructor(**entity, client=self._client)

        return entity

    def list(
        self, offset: int, limit: int, query_params: Optional[dict[str, str]] = None
    ) -> tuple[list[T], int]:
        query_params = query_params or {}
        query_params |= {
            "offset": str(offset),
            "limit": str(limit),
            "archived": "false",
            "sortBy[createdAt]": "desc",
        }

        response = self._client.get(self._base_url, query_params=query_params)
        total = int(response.headers.get("x-total-count"))
        entities = [
            self._constructor(**entity, client=self._client)
            for entity in response.json()
        ]

        return entities, total


class Result(Generic[T]):
    """Utility class for handling paginated API results"""

    def __init__(
        self,
        limit: int,
        query_params: Optional[dict[str, str]],
        requests: CRUDClient[T],
        offset: int = 0,
    ):
        self._data = deque()
        self.limit = limit
        self.offset = offset
        self._total = None
        self._query_params = query_params
        self._requests = requests

    def __iter__(self) -> "Result[T]":
        return self

    def __next__(self) -> T:
        if not self._data:
            self._fetch_next()
        return self._data.popleft()

    def __len__(self) -> int:
        if self._total is None:
            _, total = self._requests.list(
                0,
                0,
                self._query_params,
            )
            self._total = total
        return self._total

    def _fetch_next(self) -> None:
        entities, total = self._requests.list(
            self.offset,
            self.limit,
            self._query_params,
        )

        self.offset += self.limit
        self._data.extend(entities)
        self._total = total

        if len(self._data) == 0:
            raise StopIteration("No results available in the API")

    def is_empty(self) -> bool:
        """Check if the result is empty"""
        return self._total == 0



"""DHL SDK for Python"""

__all__ = [
    "APIKeyAuthentication",
    "DataHowLabClient",
    "Product",
    "Variable",
    "Experiment",
    "Recipe",
    "VariableCategorical",
    "VariableNumeric",
    "VariableLogical",
    "VariableFlow",
    "FlowVariableReference",
    "VariableSpectrum",
    "VariableSpectrumXAxis",
    "VariableSpectrumYAxis",
]

from dhl_sdk.authentication import APIKeyAuthentication
from dhl_sdk.client import DataHowLabClient
from dhl_sdk.db_entities import (
    Experiment,
    Product,
    Recipe,
    Variable,
    VariableCategorical,
    VariableFlow,
    FlowVariableReference,
    VariableLogical,
    VariableNumeric,
    VariableSpectrum,
    VariableSpectrumXAxis,
    VariableSpectrumYAxis,
)



"""This module contains utility functions used for spectra 
validation and formatting in the SDK
"""

from typing import Optional, Protocol, Union

import numpy as np

from dhl_sdk._utils import Instance, PredictRequest
from dhl_sdk.exceptions import InvalidSpectraException

# Type Aliases
SpectraData = Union[
    list[list[float]],
    np.ndarray[tuple[int, int], Union[np.dtype[np.float32], np.dtype[np.float64]]],
]


class Dataset(Protocol):
    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring
    @property
    def variables(self) -> list:
        ...

    def get_spectrum_index(self) -> int:
        ...


class SpectraModel(Protocol):
    # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring
    @property
    def inputs(self) -> list[str]:
        ...

    @property
    def dataset(self) -> Dataset:
        ...

    @property
    def spectra_size(self) -> int:
        ...


def _validate_spectra_format(spectra: SpectraData) -> list[list[float]]:
    """
    Validates and formats the spectra.

    Parameters
    ----------
    spectra : list or numpy.ndarray
        The spectra to be validated.

    Returns
    -------
    list
        The spectra as a list.

    Raises
    ------
    InvalidSpectraException
        If the spectra is not a list or numpy array.
    """

    if isinstance(spectra, np.ndarray):
        spectra = spectra.tolist()

    elif isinstance(spectra, list):
        pass
    else:
        raise InvalidSpectraException(
            f"Spectra must be a list or numpy array, but got {type(spectra)}"
        )

    return spectra


def _convert_to_request(
    spectra: SpectraData,
    model: SpectraModel,
    inputs: Optional[dict] = None,
    batch_size: int = 50,
) -> list[dict]:
    """
    Convert spectra and inputs to a list of JSON requests.
    Handles big requests by paginating them.

    Parameters
    ----------
    spectra : list[list[float]]
        A list of spectra, where each spectrum is a list of floats.
    model : Model
        The model to use for prediction.
    inputs : dict, optional
        A dictionary of input variables and their values, by default None.
    batch_size : int, optional
        The maximum number of spectra to include in each request, by default 50.

    Returns
    -------
    list[dict]
        A list of JSON requests, where each request contains a list of instances.
    """

    # get number of vars in model from config
    variables = model.dataset.variables
    n_vars = len(variables)
    spectrum_index = model.dataset.get_spectrum_index()

    request_data = []
    # handle pagination
    for i in range(0, len(spectra), batch_size):
        instance = [None] * n_vars
        instance[spectrum_index] = Instance(values=spectra[i : i + batch_size])

        if inputs is not None:
            for input_id, input_values in inputs.items():
                for index, variable in enumerate(variables):
                    if variable.id == input_id:
                        instance[index] = Instance(
                            values=input_values[i : i + batch_size]
                        )
                        break

        json_data = PredictRequest(instances=[instance]).model_dump(by_alias=True)
        request_data.append(json_data)

    return request_data
